-- Author : Nikunj Mehta
-- Date : 2/5/2002
-- A structural model of Alfa in terms of state changes

module systems/alfa-static

-- Define a portal, and the relation jumper that connects it to possibly one other portal
sig Portal {} 

-- Define input and output portals to be a partition of Portal
part sig Input, Output extends Portal {} 

-- Define particles 
sig Particle {} 

-- Define transports and fragments to be a partition of Particle 
part sig Transport, Fragment extends Particle {}

sig State {
	particles: set Particle,

	// link.Particle - gives starting points of links
	
	// Particle::link and link[Particle] - gives all Particle's links
	
	link: Particle -> Particle,
	
	// jumper.Input - gives outputs that have a jumper
	
	// jumper[Output] and Output::jumper gives all end points of jumpers
	
	jumper: Output !->! Input,
	
	// output.Output identifies Particles that have outputs
	
	// output[Particle] and Particle::output - gives all Particle's outputs 
	
	output: Particle !-> Output,
	
	// input.Input identifies Particles that have inputs
	
	// input[Particle] and Particle::input - gives all Particle's inputs 
	
	input: Particle !-> Input
} {
	// No particle can be linked to itself
	no p: Particle | p in p.link
	
	// Transports can be linked only to Fragments and vice versa
	all p: Particle | p in Transport <=> link[p] in Fragment
	//all p: Particle | p in Fragment <=> link[p] in Transport
	
	// The graph of particles should be a connected one
	some p: link.Particle | p in Particle::*link
	
	// A link is obtained from the associated portals that have a jumper relating them
	all p: Particle | link[p] = input.(jumper[output[p]])
	
	// All starting and ending points of links are particles in this state
	all p: particles | link.p + link[p] in particles
	
	// For any instance of the model, there should be some particle
	some particles
	
	// a particle can have at most one direct link to another particle 
	all p: particles | all a, b : output[p] | input.(jumper[a]) = input.(jumper[b]) => a = b
}

fun State::LinkExists(from, to: Particle) {
	(from -> to) in this.link
}

fun someState() {some s:State | some s.link}

fun cycleThrice() {some s: State | some p: s.particles | p in s.link[s.link[s.link[p]]]}

fun cycleTwice() {some s: State | some p: s.particles | p in s.link[s.link[p]]}

fun simulateCreate() {some s, s':State | one disj a, b: s.particles | s..CreateLink(s', a, b)}

assert Complementary {all s: State | all p: s.link.Particle | p in Fragment <=> s.link[p] in Transport}

fun extraParticles() {
	some p: Particle | some s: State | p !in s.particles
}

run extraParticles for 3

check Complementary for 5

run cycleThrice for 7

run cycleTwice for 4 but 1State

run simulateCreate for 6

run someState for 2